
;;;======================================================
;;;   Farmer's Dilemma Problem
;;;
;;;     Another classic AI problem (cannibals and the 
;;;     missionary) in agricultural terms. The point is
;;;     to get the farmer, the fox the cabbage and the
;;;     goat across a stream.
;;;     But the boat only holds 2 items. If left 
;;;     alone with the goat, the fox will eat it. If
;;;     left alone with the cabbage, the goat will eat
;;;     it.
;;;
;;;     To execute, merely load, reset and run.
;;;======================================================

;;;****************************
;;;* Farmer's Dilemma Problem *
;;;****************************

;;; The status facts hold the state information of the search tree.
;;; (status <search-depth>
;;;         <id>
;;;         <parent>
;;;         <farmer-location>
;;;         <fox-location>
;;;         <goat-location>
;;;         <cabbage-location>)

;;;*****************
;;;* Initial State *
;;;*****************

(deffacts initial-positions
  (status 1 initial-setup no-parent shore-1 shore-1 shore-1 shore-1))

(deffacts opposites
  (opposite-of shore-1 shore-2)
  (opposite-of shore-2 shore-1))

;;;************************
;;;* Generate Paths Rules *
;;;************************

(defrule move-alone ""
  (status ?num ?name ? ?fs ?xs ?gs ?cs)
  (opposite-of ?fs ?ns)
  =>
  (assert (status =(+ 1 ?num) =(gensym) ?name ?ns ?xs ?gs ?cs)))

(defrule move-with-fox ""
  (status ?num ?name ? ?fs ?fs ?gs ?cs)
  (opposite-of ?fs ?ns)
  =>
  (assert (status =(+ 1 ?num) =(gensym) ?name ?ns ?ns ?gs ?cs)))

(defrule move-with-goat ""
  (status ?num ?name ? ?fs ?xs ?fs ?cs)
  (opposite-of ?fs ?ns)
  =>
  (assert (status =(+ 1 ?num) =(gensym) ?name ?ns ?xs ?ns ?cs)))

(defrule move-with-cabbage ""
  (status ?num ?name ? ?fs ?xs ?gs ?fs)
  (opposite-of ?fs ?ns)
  =>
  (assert (status =(+ 1 ?num) =(gensym) ?name ?ns ?xs ?gs ?ns)))

;;;******************************
;;;* Constraint Violation Rules *
;;;******************************

(defrule fox-eats-goat ""
  (declare (salience 10000))
  ?rm <- (status ? ? ? ?s1 ?s2 ?s2 ?)
  (opposite-of ?s1 ?s2)
  =>
  (retract ?rm))

(defrule goat-eats-cabbage ""
  (declare (salience 10000))
  ?rm <- (status ? ? ? ?s1 ? ?s2 ?s2)
  (opposite-of ?s1 ?s2)
  =>
  (retract ?rm))

(defrule circular-path ""
  (declare (salience 10000))
  (status ?nm ?name ? ?fs ?xs ?gs ?cs)
  ?rm <- (status ?nm1&:(< ?nm ?nm1) ? ? ?fs ?xs ?gs ?cs)
  =>
  (retract ?rm))

;;;*********************************
;;;* Find and Print Solution Rules *
;;;*********************************

(defrule find-solution ""
  (declare (salience 5000))
  ?sol <- (status ?num ?name ?parent shore-2 shore-2 shore-2 shore-2)
  =>
  (retract ?sol)
  (printout crlf "Solution found: " crlf crlf)
  (assert (print-depth 1))
  (assert (find-move ?num ?name ?parent shore-2 shore-2 shore-2 shore-2))
  (assert (solution ?num ?name ?parent shore-2 shore-2 shore-2 shore-2)))

(defrule produce-print-path ""
  (declare (salience 6000))
  (find-move ?num ?name ?parent ?fs ?xs ?gs ?cs)
  (status ?last ?parent ?np ?fs1 ?xs1 ?gs1 ?cs1)
  =>
  (if (! (eq ?xs ?xs1))
      then
      (assert (farmer-move ?last fox ?fs))
      else
      (if (! (eq ?gs ?gs1))
	  then
	  (assert (farmer-move ?last goat ?fs))
	  else
	  (if (! (eq ?cs ?cs1))
	      then
	      (assert (farmer-move ?last cabbage ?fs))
	      else
	      (assert (farmer-move ?last alone ?fs)))))
  (assert (find-move ?last ?parent ?np ?fs1 ?xs1 ?gs1 ?cs1)))

(defrule print-it ""
  (declare (salience 5900))
  ?rm1 <- (print-depth ?depth)
  ?rm2 <- (farmer-move ?depth ?thing ?shore)
  =>
  (retract ?rm1 ?rm2)
  (if (eq ?thing alone)
      then
      (printout "Farmer moves alone to " ?shore "." crlf)
      else
      (printout "Farmer moves with " ?thing " to " ?shore "." crlf))
  (assert (print-depth =(+ 1 ?depth))))

(defrule cleanup-1 ""
  (declare (salience 5800))
  ?rm <- (find-move ?num ?name ?parent ? ? ? ?)
  =>
  (retract ?rm))

(defrule cleanup-2 ""
  (declare (salience 5800))
  ?rm <- (print-depth ?)
  =>
  (retract ?rm))
