
; This is the source code for the NASA CLIPS Inference Engine implementation
; of Conrad's "life."  The purpose of this code is to benchmark the relative
; performance of the CLIPS system to OPS5 on the VAX.  This file created on
; 9 October 1985 by D. M. Skapura.


(deffacts initial-generation
 
  (cell 1 1 0 0 0 no)
  (cell 2 1 1 0 0 no)
  (cell 3 1 0 0 0 no)
  (cell 4 1 0 0 0 no)
  (cell 5 1 1 0 0 no)

  (cell 1 2 0 0 0 no)
  (cell 2 2 1 0 0 no)
  (cell 3 2 1 0 0 no)
  (cell 4 2 1 0 0 no)
  (cell 5 2 1 0 0 no)

  (cell 1 3 0 0 0 no)
  (cell 2 3 1 0 0 no)
  (cell 3 3 0 0 0 no)
  (cell 4 3 0 0 0 no)
  (cell 5 3 1 0 0 no)

  (cell 1 4 0 0 0 no)
  (cell 2 4 0 0 0 no)
  (cell 3 4 0 0 0 no)
  (cell 4 4 1 0 0 no)
  (cell 5 4 0 0 0 no)

  (cell 1 5 0 0 0 no)
  (cell 2 5 0 0 0 no)
  (cell 3 5 0 0 0 no)
  (cell 4 5 0 0 0 no)
  (cell 5 5 0 0 0 no)

  (dimensions 5 5)
  (last-gen 4 ))


(defrule life-init
  ?start <- (initial-fact)
  =>
  (assert (cur-gen 0))
  (assert (task print-life))
  (printout "Beginning Life." crlf)
  (retract ?start))

;;;*************************
;;;* PRINT-LIFE TASK RULES *
;;;*************************

(defrule start-print
  (task print-life)
  (cur-gen ?g)
  =>
  (assert (print-row 1))
  (assert (print-col 1))
  (printout "Generation " ?g crlf))


(defrule living-print 
  (dimensions ?dx ?dy)
  ?a <- (print-col ?xx&:(<= ?xx ?dx))
  (print-row ?yy&:(<= ?yy ?dy))
  (cell ?xx ?yy 1 ? ? ?)
  =>
  (printout "x")
  (retract ?a)
  (assert (print-col =(+ 1 ?xx))))


(defrule dead-print 
  (dimensions ?dx ?dy)
  ?a <- (print-col ?xx&:(<= ?xx ?dx))
  (print-row ?yy&:(<= ?yy ?dy))
  (cell ?xx ?yy 0 ? ? ?)
  =>
  (printout "o")
  (retract ?a)
  (assert (print-col =(+ 1 ?xx))))


(defrule inc-row
  ?b <- (print-row ?yy)
  (dimensions ?dx ?)
  ?a <- (print-col ?xx&:(> ?xx ?dx))
  =>
  (printout crlf)
  (retract ?a ?b)
  (assert (print-row =(+ 1 ?yy)))
  (assert (print-col 1)))


(defrule cleanup-print
  ?c <- (task print-life)
  (last-gen ?gfinal)
  (cur-gen ?g&:(< ?g ?gfinal))
  (dimensions ?dx ?dy)
  ?a <- (print-col ?xx)
  ?b <- (print-row ?yy&:(> ?yy ?dy))
  =>
  (retract ?a ?b ?c)
  (printout "Making Neighbors." crlf)
  (assert (task do-neighbors)))

;;;***************************
;;;* DO-NEIGHBORS TASK RULES *
;;;***************************

(defrule make-neighbors
  (task do-neighbors)
  ?a <- (cell ?xvalue ?yvalue 1 ?neighbors ?generation no)
  =>
  (retract ?a)
  (assert (cell ?xvalue ?yvalue 1 ?neighbors ?generation yes))

  (assert (neighbor =(- ?xvalue 1) =(- ?yvalue 1)))
  (assert (neighbor =(- ?xvalue 1) ?yvalue))
  (assert (neighbor =(- ?xvalue 1) =(+ 1 ?yvalue)))

  (assert (neighbor ?xvalue  =(- ?yvalue 1)))
  (assert (neighbor ?xvalue  =(+ 1 ?yvalue)))

  (assert (neighbor =(+ 1 ?xvalue) =(- ?yvalue 1)))
  (assert (neighbor =(+ 1 ?xvalue) ?yvalue))
  (assert (neighbor =(+ 1 ?xvalue) =(+ 1 ?yvalue))))


(defrule end-do-neighbors
  (declare (salience -1))
  ?a <- (task do-neighbors)
  (~neighbor ? ?)
  =>
  (retract ?a)
  (printout "Computing Next Generation." crlf)
  (assert (task compute-next-generation)))


(defrule compute-neighbors
  (declare (salience 20))
  (task do-neighbors)
  ?a <- (cell ?xvalue ?yvalue ?alive ?n ?generation ?n-made)
  ?b <- (neighbor ?xvalue ?yvalue)
  =>
  (retract ?a ?b)
  (assert (cell ?xvalue ?yvalue ?alive =(+ 1 ?n) ?generation ?n-made)))


(defrule cleanup-neighbors
  (task do-neighbors)
  ?a <- (neighbor ?x ?y)
  (~cell ?x ?y ? ? ? ?)
  =>
  (retract ?a))

;;;**************************************
;;;* COMPUTE-NEXT-GENERATION TASK RULES *
;;;**************************************

(defrule continue-life
  (task compute-next-generation)
  (cur-gen ?g)
  ?a <- (cell ?xvalue ?yvalue 1 2|3 ?g ?)
  =>
  (retract ?a)
  (assert (cell ?xvalue ?yvalue 1 0 =(+ 1 ?g) no)))


(defrule die-crowded-or-lonely
  (task compute-next-generation)
  (cur-gen ?g)
  ?a <- (cell ?xvalue ?yvalue 1 ~2&~3 ?g ?)
  =>
  (retract ?a)
  (assert (cell ?xvalue ?yvalue 0 0 =(+ 1 ?g) no)))


(defrule new-life
  (task compute-next-generation)
  (cur-gen ?g)
  ?a <- (cell ?xvalue ?yvalue 0 3 ?g ?)
  =>
  (retract ?a)
  (assert (cell ?xvalue ?yvalue 1 0 =(+ 1 ?g) no)))


(defrule update-generation
  (task compute-next-generation)
  (cur-gen ?g)
  ?a <- (cell ?xvalue ?yvalue 0 ~3 ?g ?)
  =>
  (retract ?a)
  (assert (cell ?xvalue ?yvalue 0 0 =(+ 1 ?g) no)))


(defrule end-comp-loop
  (declare (salience -5))
  ?a <- (task compute-next-generation)
  ?b <- (cur-gen ?g)
  =>
  (retract ?a ?b)
  (assert (cur-gen =(+ 1 ?g)))
  (assert (task print-life)))


(defrule end-life
  (declare (salience -5))
  ?a <- (cur-gen ?g)
  (last-gen ?final&:(< ?final ?g))
  =>
  (retract ?a)
  (printout "Ending Life."))
